@@ -1,32 +1,34 @@
 from __future__ import absolute_import
 
+import collections
+import opcode
 import os
 import threading
 from collections import defaultdict
+from itertools import islice
 from os import getpid
 
-from colorama import AnsiToWin32
-
 from . import config
 from .util import BUILTIN_SYMBOLS
 from .util import CALL_COLORS
 from .util import CODE_COLORS
 from .util import MISSING
 from .util import OTHER_COLORS
-from .util import STRING_TYPES
+from .util import PY3
+from .util import StringType
 from .util import builtins
+from .util import frame_iterator
 from .util import iter_symbols
 from .util import safe_repr
+from colorama import AnsiToWin32
 
 try:
     from threading import get_ident
 except ImportError:
     from thread import get_ident
 
-
 __all__ = ['Action', 'Debugger', 'Manhole', 'CodePrinter', 'CallPrinter', 'VarsPrinter']
 
-
 BUILTIN_REPR_FUNCS = {
     'repr': repr,
     'safe_repr': safe_repr
@@ -38,11 +40,23 @@ class Action(object):
         raise NotImplementedError()
 
 
+class LazyImportPdb(object):
+    def __repr__(self):
+        return "<class 'pdb.Pdb'>"
+
+    __str__ = __repr__
+
+    def __call__(self, **kwargs):
+        from pdb import Pdb
+        return Pdb(**kwargs)
+
+
 class Debugger(Action):
     """
     An action that starts ``pdb``.
     """
-    def __init__(self, klass=config.Default('klass', lambda **kwargs: __import__('pdb').Pdb(**kwargs)), **kwargs):
+
+    def __init__(self, klass=config.Default('klass', LazyImportPdb()), **kwargs):
         self.klass = config.resolve(klass)
         self.kwargs = kwargs
 
@@ -108,9 +122,10 @@ class ColorStreamAction(Action):
                  repr_func=config.Default('repr_func', 'safe_repr')):
         self.force_colors = config.resolve(force_colors)
         self.force_pid = config.resolve(force_pid)
-        self.stream = config.resolve(stream)
-        if self.stream is None:
-            self.stream = config.DEFAULT_STREAM
+        stream = config.resolve(stream)
+        if stream is None:
+            from . import _default_stream as stream
+        self.stream = stream
         self.filename_alignment = config.resolve(filename_alignment)
         self.thread_alignment = config.resolve(thread_alignment)
         self.pid_alignment = config.resolve(pid_alignment)
@@ -149,7 +164,7 @@ class ColorStreamAction(Action):
 
     @stream.setter
     def stream(self, value):
-        if isinstance(value, STRING_TYPES):
+        if isinstance(value, StringType):
             if value in self._stream_cache:
                 value = self._stream_cache[value]
             else:
@@ -180,6 +195,27 @@ class ColorStreamAction(Action):
         else:
             raise TypeError('Expected a callable or either "repr" or "safe_repr" strings, not {!r}.'.format(value))
 
+    def try_str(self, obj):
+        """
+        Safely call ``str(obj)``. Failures will have special colored output and output is trimmed according
+        to ``self.repr_limit``.
+
+        Only used when dumping detached events.
+
+        Returns: string
+        """
+        limit = self.repr_limit
+        try:
+            s = str(obj)
+            s = s.replace('\n', r'\n')
+            if len(s) > limit:
+                cutoff = limit // 2
+                return '{} {CONT}[...]{RESET} {}'.format(s[:cutoff], s[-cutoff:], **self.other_colors)
+            else:
+                return s
+        except Exception as exc:
+            return '{INTERNAL-FAILURE}!!! FAILED REPR: {INTERNAL-DETAIL}{!r}{RESET}'.format(exc, **self.other_colors)
+
     def try_repr(self, obj):
         """
         Safely call ``self.repr_func(obj)``. Failures will have special colored output and output is trimmed according
@@ -208,7 +244,7 @@ class ColorStreamAction(Action):
         """
         source = event.fullsource if full else event.source
         if source.startswith('??? NO SOURCE: '):
-            return '{SOURCE-FAILURE}??? NO SOURCE: {SOURCE-DETAIL}{}'.format(source[15:], **self.other_colors),
+            return '{SOURCE-FAILURE}??? NO SOURCE: {SOURCE-DETAIL}{}'.format(source[15:], **self.other_colors)
         elif source:
             return source
         else:
@@ -222,11 +258,16 @@ class ColorStreamAction(Action):
         Returns: string
         """
         if event:
-            filename = event.filename or '<???>'
+            if event.builtin:
+                filename = '<{}>'.format(event.module)
+                lineno = '      '
+            else:
+                filename = event.filename or '<???>'
+                lineno = '{COLON}:{LINENO}{:<5}'.format(event.lineno, **self.other_colors)
             if len(filename) > self.filename_alignment:
                 filename = '[...]{}'.format(filename[5 - self.filename_alignment:])
-            return '{:>{}}{COLON}:{LINENO}{:<5} '.format(
-                filename, self.filename_alignment, event.lineno, **self.other_colors)
+            return '{:>{}}{} '.format(
+                filename, self.filename_alignment, lineno, **self.other_colors)
         else:
             return '{:>{}}       '.format('', self.filename_alignment)
 
@@ -298,6 +339,9 @@ class ColorStreamAction(Action):
             **dict(self.other_colors, **kwargs)
         ))
 
+    def cleanup(self):
+        pass
+
 
 class CodePrinter(ColorStreamAction):
     """
@@ -361,7 +405,7 @@ class CodePrinter(ColorStreamAction):
                 self.filename_prefix(),
                 '...',
                 event.kind,
-                event.arg if event.detached else self.try_repr(event.arg),
+                self.try_str(event.arg) if event.detached else self.try_repr(event.arg),
                 COLOR=self.event_colors.get(event.kind),
             )
 
@@ -383,9 +427,14 @@ class CallPrinter(CodePrinter):
     """
     EVENT_COLORS = CALL_COLORS
 
+    locals = defaultdict(list)
+
+    @classmethod
+    def cleanup(cls):
+        cls.locals = defaultdict(list)
+
     def __init__(self, *args, **kwargs):
         super(CallPrinter, self).__init__(*args, **kwargs)
-        self.locals = defaultdict(list)
 
     def __call__(self, event):
         """
@@ -401,7 +450,20 @@ class CallPrinter(CodePrinter):
         thread_prefix = self.thread_prefix(event)
         filename_prefix = self.filename_prefix(event)
 
-        if event.kind == 'call':
+        if event.builtin and event.kind == 'call':
+            stack.append(ident)
+            self.output(
+                '{}{}{}{KIND}{:9} {}{COLOR}=>{NORMAL} {}: {}{RESET}\n',
+                pid_prefix,
+                thread_prefix,
+                filename_prefix,
+                event.kind,
+                '   ' * (len(stack) - 1),
+                event.function,
+                self.try_source(event).strip(),
+                COLOR=self.event_colors.get(event.kind),
+            )
+        elif event.kind == 'call':
             code = event.code
             stack.append(ident)
             self.output(
@@ -414,7 +476,7 @@ class CallPrinter(CodePrinter):
                 event.function,
                 ', '.join('{VARS}{VARS-NAME}{0}{VARS}={RESET}{1}'.format(
                     var,
-                    event.locals.get(var, MISSING) if event.detached else self.try_repr(event.locals.get(var, MISSING)),
+                    self.try_str(event.locals.get(var, MISSING)) if event.detached else self.try_repr(event.locals.get(var, MISSING)),
                     **self.other_colors
                 ) for var in code.co_varnames[:code.co_argcount]),
                 COLOR=self.event_colors.get(event.kind),
@@ -428,20 +490,21 @@ class CallPrinter(CodePrinter):
                 event.kind,
                 '   ' * (len(stack) - 1),
                 event.function,
-                event.arg if event.detached else self.try_repr(event.arg),
+                self.try_str(event.arg) if event.detached else self.try_repr(event.arg),
                 COLOR=self.event_colors.get(event.kind),
             )
 
         elif event.kind == 'return':
             self.output(
-                '{}{}{}{KIND}{:9} {}{COLOR}<={NORMAL} {}: {RESET}{}\n',
+                '{}{}{}{KIND}{:9} {}{COLOR}<={NORMAL} {}{}{RESET}{}\n',
                 pid_prefix,
                 thread_prefix,
                 filename_prefix,
                 event.kind,
                 '   ' * (len(stack) - 1),
                 event.function,
-                event.arg if event.detached else self.try_repr(event.arg),
+                '' if event.builtin else ': ',
+                '' if event.builtin else (self.try_str(event.arg) if event.detached else self.try_repr(event.arg)),
                 COLOR=self.event_colors.get(event.kind),
             )
             if stack and stack[-1] == ident:
@@ -500,12 +563,12 @@ class VarsPrinter(ColorStreamAction):
         for code, symbols in self.names.items():
             try:
                 obj = eval(code, dict(vars(builtins), **event.globals), event.locals)
-            except AttributeError:
+            except (AttributeError, KeyError):
                 continue
             except Exception as exc:
                 printout = '{INTERNAL-FAILURE}FAILED EVAL: {INTERNAL-DETAIL}{!r}'.format(exc, **self.other_colors)
             else:
-                printout = obj if event.detached else self.try_repr(obj)
+                printout = self.try_str(obj) if event.detached else self.try_repr(obj)
 
             if frame_symbols >= symbols:
                 if first:
@@ -534,12 +597,21 @@ class VarsSnooper(ColorStreamAction):
     """
     A PySnooper-inspired action, similar to :class:`~hunter.actions.VarsPrinter`, but only show variable changes.
 
-    .. warning: Should be considered experimental. Use judiciously.
+    .. warning:: Should be considered experimental. Use judiciously.
 
         * It stores reprs for all seen variables, therefore it can use lots of memory.
         * Will leak memory if you filter the return events (eg: ``~Q(kind="return")``).
         * Not thoroughly tested. May misbehave on code with closures/nonlocal variables.
+
+    Args:
+        stream (file-like): Stream to write to. Default: ``sys.stderr``.
+        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
+        force_colors (bool): Force coloring. Default: ``False``.
+        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
+        repr_func (string or callable): Function to use instead of ``repr``.
+            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.
     """
+
     def __init__(self, **options):
         super(VarsSnooper, self).__init__(**options)
         self.stored_reprs = defaultdict(dict)
@@ -556,7 +628,7 @@ class VarsSnooper(ColorStreamAction):
         empty_filename_prefix = self.filename_prefix()
 
         current_reprs = {
-            name: value if event.detached else self.try_repr(value)
+            name: self.try_str(value) if event.detached else self.try_repr(value)
             for name, value in event.locals.items()
         }
         scope_key = event.code or event.function
@@ -613,3 +685,153 @@ class VarsSnooper(ColorStreamAction):
                     )
         if event.kind == 'return':
             del self.stored_reprs[scope_key]
+
+
+RETURN_VALUE = opcode.opmap['RETURN_VALUE']
+if not PY3:
+    RETURN_VALUE = chr(RETURN_VALUE)
+
+
+class ErrorSnooper(CodePrinter):
+    """
+    An action that prints events around silenced exceptions. Note that it inherits the output of :class:`~hunter.actions.CodePrinter` so no
+    fancy call indentation.
+
+    .. warning::
+
+        Should be considered experimental. May show lots of false positives especially if you're tracing lots of clumsy code like:
+
+        .. sourcecode:: python
+
+            try:
+                stuff = something[key]
+            except KeyError:
+                stuff = "default"
+
+    Args:
+        max_backlog (int): Maximum number of events to record and display before the silenced exception is raised.
+            Set to 0 to disable and get a speed boost. Default: ``10``.
+        max_events (int): Maximum number of events to record and display for each detected silenced exception. Default: ``50``.
+        max_depth (int): Increase if you want to drill into subsequent calls after an exception is raised. If you increase this you might
+            want to also increase ``max_events`` since subsequent calls may have so many events you won't get to see the return event.
+            Default: ``0`` (doesn't drill into any calls).
+
+        stream (file-like): Stream to write to. Default: ``sys.stderr``.
+        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
+        force_colors (bool): Force coloring. Default: ``False``.
+        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
+        repr_func (string or callable): Function to use instead of ``repr``.
+            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.
+
+    .. versionadded:: 3.1.0
+    """
+
+    def __init__(self, *args, **kwargs):
+        self.max_backlog = kwargs.pop('max_backlog', 10)
+        self.backlog = collections.deque(maxlen=self.max_backlog)
+        self.max_events = kwargs.pop('max_events', 50)
+        self.max_depth = kwargs.pop('max_depth', 0)
+        self.origin = None
+        self.events = None
+        super(ErrorSnooper, self).__init__(*args, **kwargs)
+
+    def __call__(self, event):
+        if self.max_backlog:
+            detached_event = event.detach(self.try_repr)
+            self.backlog.append(detached_event)
+        else:
+            detached_event = None
+
+        if event.kind == 'exception':  # something interesting happened ;)
+            if detached_event is None:
+                detached_event = event.detach(self.try_repr)
+            if self.origin:
+                self.events.append(detached_event)
+                if self.origin.depth > event.depth:
+                    self.origin = detached_event
+            else:
+                self.events = list(self.backlog)
+                self.origin = detached_event
+        elif self.origin:
+            if event.kind == 'return':
+                if detached_event is None:
+                    detached_event = event.detach(self.try_repr)
+                self.events.append(detached_event)
+                if event.depth == self.origin.depth - 1:  # stop if the same function returned (depth is -1)
+                    if event.instruction == RETURN_VALUE:
+                        self.dump_events()
+                        self.output('{BRIGHT}{fore(BLACK)}{} function exit{RESET}\n', '-' * 46)
+                    self.origin = None
+                    self.events = None
+            elif event.depth > self.origin.depth + self.max_depth:  # too many details
+                return
+            elif len(self.events) > self.max_events:
+                self.dump_events()
+                self.output('{BRIGHT}{fore(BLACK)}{} too many lines{RESET}\n', '-' * 46)
+            else:
+                if detached_event is None:
+                    detached_event = event.detach(self.try_repr)
+                self.events.append(detached_event)
+
+    def dump_events(self):
+        self.output('{BRIGHT}{fore(BLUE)}{} tracing {fore(YELLOW)}{}{fore(BLUE)} on {fore(RED)}{}{RESET}\n',
+                    '>' * 46, self.origin.function, self.origin.arg)
+        for event in self.events:
+            super(ErrorSnooper, self).__call__(event)
+        self.origin = None
+        self.events = None
+
+
+class StackPrinter(ColorStreamAction):
+    """
+    An action that prints a one-line stacktrace.
+
+    Args:
+        depth (int): The maximum number of frames to show.
+        limit (int): The maximum number of components to show in path. Eg: ``limit=2`` means it will show 1 parent: ``foo/bar.py``.
+        stream (file-like): Stream to write to. Default: ``sys.stderr``.
+        filename_alignment (int): Default size for the filename column (files are right-aligned). Default: ``40``.
+        force_colors (bool): Force coloring. Default: ``False``.
+        repr_limit (bool): Limit length of ``repr()`` output. Default: ``512``.
+        repr_func (string or callable): Function to use instead of ``repr``.
+            If string must be one of 'repr' or 'safe_repr'. Default: ``'safe_repr'``.
+    """
+
+    def __init__(self, depth=15, limit=2, **options):
+        self.limit = limit
+        self.depth = depth
+        super(StackPrinter, self).__init__(**options)
+
+    def __call__(self, event):
+        """
+        Handle event and print the stack.
+        """
+        pid_prefix = self.pid_prefix()
+        thread_prefix = self.thread_prefix(event)
+        filename_prefix = self.filename_prefix(event).rstrip()
+        sep = os.path.sep
+
+        if event.frame and event.frame.f_back:
+            template = '{}{}{}{CONT}:{BRIGHT}{fore(BLUE)}%s {KIND}<={RESET} %s' % (
+                event.function,
+                ' {KIND}<={RESET} '.join(
+                    '%s{CONT}:{RESET}%s{CONT}:{BRIGHT}{fore(BLUE)}%s' % (
+                        '/'.join(frame.f_code.co_filename.split(sep)[-self.limit:]),
+                        frame.f_lineno,
+                        frame.f_code.co_name
+                    )
+                    for frame in islice(frame_iterator(event.frame.f_back), self.depth)
+                )
+            )
+        else:
+            template = '{}{}{}{CONT}:{BRIGHT}{fore(BLUE)}%s {KIND}<= {BRIGHT}{fore(YELLOW)}no frames available {NORMAL}(detached=%s)' % (
+                event.function,
+                event.detached,
+            )
+        template += '{RESET}\n'
+        self.output(
+            template,
+            pid_prefix,
+            thread_prefix,
+            filename_prefix,
+        )
