@@ -1,9 +1,11 @@
 import ast
-import io
 import re
 import sys
 import types
 import weakref
+from collections import Counter
+from collections import OrderedDict
+from collections import defaultdict
 from collections import deque
 
 from colorama import Back
@@ -24,18 +26,24 @@ try:
     from threading import main_thread
 except ImportError:
     from threading import _shutdown
+
     get_main_thread = weakref.ref(
         _shutdown.__self__ if hasattr(_shutdown, '__self__') else _shutdown.im_self)
     del _shutdown
 else:
     get_main_thread = weakref.ref(main_thread())
 
+try:
+    from types import InstanceType
+except ImportError:
+    InstanceType = object()
+
 PY3 = sys.version_info[0] == 3
 
 if PY3:
-    STRING_TYPES = str,
+    StringType = str,
 else:
-    STRING_TYPES = basestring,  # noqa
+    StringType = basestring,  # noqa
 
 OTHER_COLORS = {
     'COLON': Style.BRIGHT + Fore.BLACK,
@@ -129,80 +137,47 @@ def iter_symbols(code):
             yield node.id
 
 
-def has_dict(obj_type, obj, tolerance=25):
-    """
-    A contrived mess to check that object doesn't have a __dit__ but avoid checking it if any ancestor is evil enough to
-    explicitly define __dict__ (like apipkg.ApiModule has __dict__ as a property).
-    """
-    ancestor_types = deque()
-    while obj_type is not type and tolerance:
-        ancestor_types.appendleft(obj_type)
-        obj_type = type(obj_type)
-        tolerance -= 1
-    for ancestor in ancestor_types:
-        __dict__ = getattr(ancestor, '__dict__', None)
-        if __dict__ is not None:
-            if '__dict__' in __dict__:
-                return True
-    return hasattr(obj, '__dict__')
-
-
 def safe_repr(obj, maxdepth=5):
     if not maxdepth:
         return '...'
     obj_type = type(obj)
+    obj_type_type = type(obj_type)
     newdepth = maxdepth - 1
 
-    # specifically handle few of the container builtins that would normally do repr on contained values
-    if isinstance(obj, dict):
-        if obj_type is not dict:
-            return '%s({%s})' % (
-                obj_type.__name__,
-                ', '.join('%s: %s' % (
-                    safe_repr(k, maxdepth),
-                    safe_repr(v, newdepth)
-                ) for k, v in obj.items()))
-        else:
-            return '{%s}' % ', '.join('%s: %s' % (
+    # only represent exact builtins
+    # (subclasses can have side-effects due to __class__ as a property, __instancecheck__, __subclasscheck__ etc)
+    if obj_type is dict:
+        return '{%s}' % ', '.join('%s: %s' % (
+            safe_repr(k, maxdepth),
+            safe_repr(v, newdepth)
+        ) for k, v in obj.items())
+    elif obj_type is list:
+        return '[%s]' % ', '.join(safe_repr(i, newdepth) for i in obj)
+    elif obj_type is tuple:
+        return '(%s%s)' % (', '.join(safe_repr(i, newdepth) for i in obj), ',' if len(obj) == 1 else '')
+    elif obj_type is set:
+        return '{%s}' % ', '.join(safe_repr(i, newdepth) for i in obj)
+    elif obj_type is frozenset:
+        return '%s({%s})' % (obj_type.__name__, ', '.join(safe_repr(i, newdepth) for i in obj))
+    elif obj_type is deque:
+        return '%s([%s])' % (obj_type.__name__, ', '.join(safe_repr(i, newdepth) for i in obj))
+    elif obj_type in (Counter, OrderedDict, defaultdict):
+        return '%s({%s})' % (
+            obj_type.__name__,
+            ', '.join('%s: %s' % (
                 safe_repr(k, maxdepth),
                 safe_repr(v, newdepth)
             ) for k, v in obj.items())
-    elif isinstance(obj, list):
-        if obj_type is not list:
-            return '%s([%s])' % (obj_type.__name__, ', '.join(safe_repr(i, newdepth) for i in obj))
-        else:
-            return '[%s]' % ', '.join(safe_repr(i, newdepth) for i in obj)
-    elif isinstance(obj, tuple):
-        if obj_type is not tuple:
-            return '%s(%s%s)' % (
-                obj_type.__name__,
-                ', '.join(safe_repr(i, newdepth) for i in obj),
-                ',' if len(obj) == 1 else '')
-        else:
-            return '(%s%s)' % (', '.join(safe_repr(i, newdepth) for i in obj), ',' if len(obj) == 1 else '')
-    elif isinstance(obj, set):
-        if obj_type is not set:
-            return '%s({%s})' % (obj_type.__name__, ', '.join(safe_repr(i, newdepth) for i in obj))
-        else:
-            return '{%s}' % ', '.join(safe_repr(i, newdepth) for i in obj)
-    elif isinstance(obj, frozenset):
-        return '%s({%s})' % (obj_type.__name__, ', '.join(safe_repr(i, newdepth) for i in obj))
-    elif isinstance(obj, deque):
-        return '%s([%s])' % (obj_type.__name__, ', '.join(safe_repr(i, newdepth) for i in obj))
-    elif isinstance(obj, BaseException):
-        return '%s(%s)' % (obj_type.__name__, ', '.join(safe_repr(i, newdepth) for i in obj.args))
-    elif isinstance(obj, types.MethodType):
+        )
+    elif obj_type is types.MethodType:  # noqa
         self = obj.__self__
         name = getattr(obj, '__qualname__', None)
         if name is None:
             name = obj.__name__
         return '<%sbound method %s of %s>' % ('un' if self is None else '', name, safe_repr(self, newdepth))
-    elif not has_dict(obj_type, obj) or obj_type in (
-        type, types.ModuleType,
-        types.FunctionType,
-        types.BuiltinFunctionType, types.BuiltinMethodType,
-        io.IOBase
-    ):
+    elif obj_type_type is type and BaseException in obj_type.__mro__:
+        return '%s(%s)' % (obj_type.__name__, ', '.join(safe_repr(i, newdepth) for i in obj.args))
+    elif obj_type_type is type and obj_type is not InstanceType and obj_type.__module__ in (builtins.__name__, 'io', 'socket', '_socket'):
         # hardcoded list of safe things. note that isinstance ain't used
         # (we don't trust subclasses to do the right thing in __repr__)
         return repr(obj)
@@ -210,3 +185,12 @@ def safe_repr(obj, maxdepth=5):
         # if the object has a __dict__ then it's probably an instance of a pure python class, assume bad things
         #  with side-effects will be going on in __repr__ - use the default instead (object.__repr__)
         return object.__repr__(obj)
+
+
+def frame_iterator(frame):
+    """
+    Yields frames till there are no more.
+    """
+    while frame:
+        yield frame
+        frame = frame.f_back
